# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 The NebulaGraph Authors. All rights reserved.

from typing import Generator, List, Dict, Literal

from nebula3.Config import Config
from nebula3.gclient.net import ConnectionPool

from ng_nx.utils import NebulaGraphConfig, result_to_df


class NebulaWriter:
    def __init__(self, data: Dict, nebula_config: NebulaGraphConfig):
        self.data = data
        self.label = None
        self.properties = []
        self.batch_size = None
        self.write_mode = None
        self.sink = None

        config = Config()
        graphd_hosts = nebula_config.graphd_hosts.split(",")
        graphd_host_list = [
            (host.split(":")[0], int(host.split(":")[1])) for host in graphd_hosts
        ]
        self.space = nebula_config.space
        self.nebula_user = nebula_config.user
        self.nebula_password = nebula_config.password
        self.connection_pool = ConnectionPool()
        assert self.connection_pool.init(
            graphd_host_list, config
        ), "Init Connection Pool Failed"

    def set_options(
        self,
        label: str,
        properties: List[str],
        batch_size: int = 256,
        write_mode: str = "insert",
        sink: Literal[
            "nebulagraph_vertex", "nebulagraph_edge"
        ] = "nebulagraph_vertex",
    ):
        self.label = label
        self.properties = properties
        self.batch_size = batch_size
        self.write_mode = write_mode.lower()
        self.sink = sink.lower()

    def write(self):
        if self.write_mode == "update":
            raise NotImplementedError("Update mode is not implemented yet")
        if self.write_mode != "insert":
            raise ValueError("Only insert mode is supported now")
        if self.sink == "nebulagraph_vertex":
            return self._write_vertex()
        elif self.sink == "nebulagraph_edge":
            return self._write_edge()
        else:
            raise ValueError("Invalid sink type")

    def _write_vertex(self):
        if self.sink == "nebulagraph_edge":
            raise NotImplementedError("Edge sink is not implemented yet")
        if self.write_mode != "insert" or self.sink != "nebulagraph_vertex":
            raise ValueError("Only insert mode and vertex sink are supported now")
        # TODO: add more write modes and sinks later
        with self.connection_pool.session_context(
            self.nebula_user, self.nebula_password
        ) as session:
            assert session.execute(
                f"USE {self.space}"
            ).is_succeeded(), f"Failed to use space {self.space}"
            # get types of properties
            properties_types = {}
            query = f"DESC TAG {self.label}"
            result = session.execute(query)
            assert result.is_succeeded(), (
                f"Failed to get types of properties: {result.error_msg()}, "
                f"consider creating TAG {self.label} first."
            )
            types_df = result_to_df(result)

            for i in range(len(types_df)):
                properties_types[types_df.iloc[i, 0]] = types_df.iloc[i, 1]
            for property in self.properties:
                if property not in properties_types:
                    raise ValueError(
                        f"Property {property} is not defined in TAG {self.label}"
                    )

            # Set up the write query in batches, string value should be enclosed by double quotes
            # INSERT VERTEX {label} ({properties}) VALUES
            # "{vid}":({value0, value1}),...;
            query_prefix = (
                f"INSERT VERTEX {self.label} ({','.join(self.properties)}) VALUES "
            )
            query = query_prefix
            # now self.data is a dict or a generator object, key is vid, value is a list of properties or single property
            quote = '"'
            if isinstance(self.data, dict):
                for i, (vid, properties) in enumerate(self.data.items()):
                    if i % self.batch_size == 0 and i != 0:
                        # execute the query
                        # remove tailing comma
                        query = query[:-1]
                        result = session.execute(query)
                        assert (
                            result.is_succeeded()
                        ), f"Failed to write data: {result.error_msg()}"
                        # reset the query
                        query = query_prefix
                    if isinstance(properties, list):
                        # should add quotes to string values
                        query += f"{quote}{vid}{quote}:({','.join([quote+str(value)+quote if properties_types[self.properties[i]] == 'string' else str(value) for i, value in enumerate(properties)])}),"
                    else:
                        query += f"{quote}{vid}{quote}:({quote+str(properties)+quote if properties_types[self.properties[0]] == 'string' else str(properties)}),"
                # execute the last query
                query = query[:-1]
                result = session.execute(query)
                assert (
                    result.is_succeeded()
                ), f"Failed to write data: {result.error_msg()}"
            elif isinstance(self.data, Generator):
                for i, (vid, properties) in enumerate(self.data):
                    if i % self.batch_size == 0 and i != 0:
                        # execute the query
                        # remove tailing comma
                        query = query[:-1]
                        result = session.execute(query)
                        assert (
                            result.is_succeeded()
                        ), f"Failed to write data: {result.error_msg()}"
                        # reset the query
                        query = query_prefix
                    if isinstance(properties, list):
                        # should add quotes to string values
                        query += f"{quote}{vid}{quote}:({','.join([quote+str(value)+quote if properties_types[self.properties[i]] == 'string' else str(value) for i, value in enumerate(properties)])}),"
                    else:
                        query += f"{quote}{vid}{quote}:({quote+str(properties)+quote if properties_types[self.properties[0]] == 'string' else str(properties)}),"
                # execute the last query
                query = query[:-1]
                result = session.execute(query)
                assert (
                    result.is_succeeded()
                ), f"Failed to write data: {result.error_msg()}"
            else:
                raise TypeError("Data should be a dict or a generator object")
        return True

    def _write_edge(self):
        with self.connection_pool.session_context(
            self.nebula_user, self.nebula_password
        ) as session:
            assert session.execute(
                f"USE {self.space}"
            ).is_succeeded(), f"Failed to use space {self.space}"

            # Get types of edge properties
            properties_types = {}
            query = f"DESC EDGE {self.label}"
            result = session.execute(query)
            assert result.is_succeeded(), (
                f"Failed to get types of properties: {result.error_msg()}, "
                f"consider creating EDGE {self.label} first."
            )
            types_df = result_to_df(result)

            for i in range(len(types_df)):
                properties_types[types_df.iloc[i, 0]] = types_df.iloc[i, 1]
            for property in self.properties:
                if property not in properties_types:
                    raise ValueError(
                        f"Property {property} is not defined in EDGE {self.label}"
                    )

            # Set up the write query in batches
            query_prefix = (
                f"INSERT EDGE {self.label} ({','.join(self.properties)}) VALUES "
            )
            query = query_prefix
            quote = '"'

            def process_edge(src, dst, rank, props):
                nonlocal query
                prop_values = ",".join(
                    [
                        f"{quote}{str(value)}{quote}"
                        if properties_types[self.properties[i]] == "string"
                        else str(value)
                        for i, value in enumerate(props)
                    ]
                )
                query += f"{quote}{src}{quote}->{quote}{dst}{quote}@{rank}:({prop_values}),"

            batch_count = 0
            for edge in self.data:
                if batch_count == self.batch_size:
                    # Execute the query
                    query = query[:-1]  # Remove trailing comma
                    result = session.execute(query)
                    assert (
                        result.is_succeeded()
                    ), f"Failed to write data: {result.error_msg()}"
                    query = query_prefix
                    batch_count = 0

                src, dst, rank, props = edge
                process_edge(src, dst, rank, props)
                batch_count += 1

            # Execute the last batch
            if batch_count > 0:
                query = query[:-1]  # Remove trailing comma
                result = session.execute(query)
                assert (
                    result.is_succeeded()
                ), f"Failed to write data: {result.error_msg()}"

        return True



# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 The NebulaGraph Authors. All rights reserved.

from __future__ import annotations

from typing import Dict

import pandas as pd
from nebula3.data.DataObject import Value, ValueWrapper
from nebula3.data.ResultSet import ResultSet

# Default Configuration
GRAPHD_HOSTS = "graphd:9669"
METAD_HOSTS = "metad0:9559,metad1:9559,metad2:9559"
USER = "root"
PASSWORD = "nebula"
SPACE = "basketballplayer"

# TBD when https://github.com/vesoft-inc/nebula-python/pull/269 is merged
# in a release version of nebula-python, I could leverage the ValueWrapper.cast()
# method to simplify the code below
cast_as = {
    Value.NVAL: "as_null",
    Value.BVAL: "as_bool",
    Value.IVAL: "as_int",
    Value.FVAL: "as_double",
    Value.SVAL: "as_string",
    Value.LVAL: "as_list",
    Value.UVAL: "as_set",
    Value.MVAL: "as_map",
    Value.TVAL: "as_time",
    Value.DVAL: "as_date",
    Value.DTVAL: "as_datetime",
    Value.VVAL: "as_node",
    Value.EVAL: "as_relationship",
    Value.PVAL: "as_path",
    Value.GGVAL: "as_geography",
    Value.DUVAL: "as_duration",
}


def cast(val: ValueWrapper):
    _type = val._value.getType()
    if _type == Value.__EMPTY__:
        return None
    if _type in cast_as:
        return getattr(val, cast_as[_type])()
    if _type == Value.LVAL:
        return [x.cast() for x in val.as_list()]
    if _type == Value.UVAL:
        return {x.cast() for x in val.as_set()}
    if _type == Value.MVAL:
        return {k: v.cast() for k, v in val.as_map().items()}


def result_to_df(result: ResultSet) -> pd.DataFrame:
    """
    build list for each column, and transform to dataframe
    """
    assert result.is_succeeded()
    columns = result.keys()
    d: Dict[str, list] = {}
    for col_num in range(result.col_size()):
        col_name = columns[col_num]
        col_list = result.column_values(col_name)
        d[col_name] = [cast(x) for x in col_list]
    return pd.DataFrame(d)


class NebulaGraphConfig:
    def __init__(
        self,
        graphd_hosts: str = GRAPHD_HOSTS,
        metad_hosts: str = METAD_HOSTS,
        user: str = USER,
        password: str = PASSWORD,
        space: str = SPACE,
        **kwargs,
    ):
        self.graphd_hosts = graphd_hosts
        self.metad_hosts = metad_hosts
        self.user = user
        self.password = password
        self.space = space
        self.shuffle_partitions = None
        self.executor_memory = None
        self.driver_memory = None
        for key, value in kwargs.items():
            setattr(self, key, value)

    def __getattr__(self, name):
        return None



# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 The NebulaGraph Authors. All rights reserved.

from typing import Dict

import networkx as nx
import pandas as pd
from nebula3.Config import Config
from nebula3.data.ResultSet import ResultSet
from nebula3.gclient.net import ConnectionPool

from ng_nx.utils import NebulaGraphConfig


class NebulaScanReader:
    pass



# SPDX-License-Identifier: Apache-2.0
# Copyright 2024 The NebulaGraph Authors. All rights reserved.

from pkgutil import extend_path

__path__ = extend_path(__path__, __name__)  # type: ignore

from ng_nx.query_reader import NebulaReader, NebulaQueryReader
from ng_nx.scan_reader import NebulaScanReader
from ng_nx.writer import NebulaWriter

# export
__all__ = (
    "NebulaReader",
    "NebulaScanReader",
    "NebulaWriter",
    "NebulaQueryReader",
)



# SPDX-License-Identifier: Apache-2.0
# Copyright 2024 The NebulaGraph Authors. All rights reserved.

import networkx as nx
import pandas as pd
from nebula3.Config import Config
from nebula3.gclient.net import ConnectionPool
from nebula3.data.ResultSet import ResultSet

from ng_nx.utils import NebulaGraphConfig, result_to_df


class NebulaReader:
    def __init__(
        self,
        edges: list,
        properties: list,
        nebula_config: NebulaGraphConfig,
        limit: int,
    ):
        self.edges = edges
        self.properties = properties
        self.limit = limit

        config = Config()
        graphd_hosts = nebula_config.graphd_hosts.split(",")
        graphd_host_list = [
            (host.split(":")[0], int(host.split(":")[1])) for host in graphd_hosts
        ]
        self.space = nebula_config.space
        self.nebula_user = nebula_config.user
        self.nebula_password = nebula_config.password
        self.connection_pool = ConnectionPool()
        assert self.connection_pool.init(
            graphd_host_list, config
        ), "Init Connection Pool Failed"
        assert len(edges) > 0 and len(edges) == len(
            properties
        ), "edges and properties should have the same length"

    def read(self):
        with self.connection_pool.session_context(
            self.nebula_user, self.nebula_password
        ) as session:
            assert session.execute(
                f"USE {self.space}"
            ).is_succeeded(), f"Failed to use space {self.space}"
            result_list = []
            g = nx.MultiDiGraph()
            for i in range(len(self.edges)):
                edge = self.edges[i]
                properties = self.properties[i]
                properties_query_field = ""
                for property in properties:
                    properties_query_field += f", e.{property} AS {property}"
                result = session.execute(
                    f"MATCH ()-[e:`{edge}`]->() RETURN src(e) AS src, dst(e) AS dst{properties_query_field} LIMIT {self.limit}"
                )
                # print(f'query: MATCH ()-[e:`{edge}`]->() RETURN src(e) AS src, dst(e) AS dst{properties_query_field} LIMIT {self.limit}')
                # print(f"Result: {result}")
                assert result.is_succeeded()
                result_list.append(result)

            # merge all result
            for i, result in enumerate(result_list):
                _df = result_to_df(result)
                # TBD, consider add label of edge
                properties = self.properties[i] if self.properties[i] else None
                _g = nx.from_pandas_edgelist(
                    _df, "src", "dst", properties, create_using=nx.MultiDiGraph()
                )
                g = nx.compose(g, _g)
            return g

    def release(self):
        self.connection_pool.close()

    def __del__(self):
        self.release()

class NebulaQueryReader:
    def __init__(self, nebula_config: NebulaGraphConfig):
        self.config = nebula_config
        self.connection_pool = ConnectionPool()
        graphd_hosts = nebula_config.graphd_hosts.split(",")
        graphd_host_list = [
            (host.split(":")[0], int(host.split(":")[1])) for host in graphd_hosts
        ]
        config = Config()
        assert self.connection_pool.init(
            graphd_host_list, config
        ), "Init Connection Pool Failed"

    def read(self, query: str) -> nx.MultiDiGraph:
        with self.connection_pool.session_context(
            self.config.user, self.config.password
        ) as session:
            assert session.execute(
                f"USE {self.config.space}"
            ).is_succeeded(), f"Failed to use space {self.config.space}"
            
            result: ResultSet = session.execute(query)
            assert result.is_succeeded(), f"Query execution failed: {result.error_msg()}"
            
            vis_data = result.dict_for_vis()
            return self._construct_graph(vis_data)

    def _construct_graph(self, vis_data: dict) -> nx.MultiDiGraph:
        g = nx.MultiDiGraph()
        
        # Add nodes
        for node_data in vis_data['nodes']:
            g.add_node(node_data['id'], **node_data['props'], labels=node_data['labels'])
        
        # Add edges
        for edge_data in vis_data['edges']:
            g.add_edge(
                edge_data['src'],
                edge_data['dst'],
                key=edge_data['name'],
                **edge_data['props']
            )
        
        return g

    def release(self):
        self.connection_pool.close()

    def __del__(self):
        self.release()